;; Definitions for Dr. Scheme
(define nil '())
(define (symbol<? symbol1 symbol2)
  (string<? (symbol->string symbol1)
            (symbol->string symbol2)))
(define (fold-right op init lst)
  (if (null? lst)
      init
      (op (car lst)
          (fold-right op init (cdr lst)))))

;; Error checking helpers
(define (assert test-val pred)
  (if (pred test-val)
      #t
      (error "ASSERT FAILED: " test-val pred)))
(define (assert-equal test-val expected-val)
  (assert test-val (lambda (t) (equal? t expected-val))))
(define (assert-f test-val)
  (assert-equal test-val #f))
(define (assert-t test-val)
  (assert-equal test-val #t))


;; Set abstraction
(define (new-set less-pred)
  (if (procedure? less-pred)
      (list 'set less-pred)
      (error "less-pred must be a procedure")))

(define (new-number-set) (new-set <))
(define (new-symbol-set) (new-set symbol<?))

(define (set? set)
  (and (list? set) 
       (not (null? set))
       (eq? (car set) 'set) 
       (procedure? (cadr set))))

(assert (new-number-set) set?)
(assert (new-symbol-set) set?)
(assert (new-set string<?) set?)

(define (set-check set)
  (if (set? set) #t (error "not a set")))

(set-check (new-number-set))
(set-check (new-symbol-set))
(set-check (new-set string<?))

(define (set-to-list set)
  (set-check set)
  (cddr set))

(assert-equal (set-to-list (new-number-set)) nil)
(assert-equal (set-to-list (new-set >)) nil)

(define (set-elm-less set)
  (set-check set)
  (cadr set))

(assert-equal (set-elm-less (new-set >)) >)

(define (set-elm-equal set)
  (let ((less (set-elm-less set)))
    (lambda (a b) (and (not (less a b))
                       (not (less b a))))))

(define (set-contains? elm set)
  (set-check set)
  (let ((less (set-elm-less set))
        (equal (set-elm-equal set)))
    (fold-right (lambda (val acc) (or acc (equal val elm)))
                #f 
                (set-to-list set))))

(assert-f (set-contains? 123 (new-number-set)))

(define (set-add elm set)
  (set-check set)
  (if (set-contains? elm set)
      set
      (append (list 'set (set-elm-less set) )
              (cons elm (set-to-list set)))))

(assert-equal (set-to-list (set-add 123 (new-number-set))) '(123))
(define s (set-add 3 (set-add 2 (set-add 1 (new-number-set)))))
(assert-t (set-contains? 1 s))
(assert-t (set-contains? 2 s))
(assert-t (set-contains? 3 s))
(assert-f (set-contains? 4 s))

(define (list-to-set less-pred lst)
  (fold-right set-add (new-set less-pred) lst))

(assert-t (set-contains? 1 (list-to-set < '(1 2 3 3))))
(assert-t (set-contains? 2 (list-to-set < '(1 2 3 3))))
(assert-t (set-contains? 3 (list-to-set < '(1 2 3 3))))
(assert-f (set-contains? 4 (list-to-set < '(1 2 3 3))))
(assert-equal (length (set-to-list (list-to-set < '(1 2 3 3)))) 3)

(define (set-remove elm set)
  (set-check set)
  (let ((less (set-elm-less set))
        (equal (set-elm-equal set)))
    (list-to-set less
                 (filter (lambda (x) (not (equal elm x)))
                         (set-to-list set)))))

(assert-equal (length (set-to-list (set-remove 3 (list-to-set < '(1 2 3 3 4 5))))) 4)
(assert-equal (length (set-to-list (set-remove 3 (set-remove 3 (list-to-set < '(1 2 3 3 4 5)))))) 4)
(assert-equal (length (set-to-list (set-remove 1 (set-remove 2 (list-to-set < '(1 2 3 3 4 5)))))) 3)
(assert-equal (length (set-to-list (set-remove 1 (set-remove 2 (set-remove 3 (set-remove 4 (set-remove 5 (list-to-set < '(1 2 3 3 4 5))))))))) 0)

;; Determines whether set 1 and set 2 contain the same
;; elements, based on s1's definition of equivalence.
(define (set-equal? s1 s2)
  (set-check s1) (set-check s2)
  (let* ((lst1 (set-to-list s1))
         (l1 (length lst1))
         (l2 (length (set-to-list s2)))
         (equal (set-elm-equal s1)))
    (if (not (= l1 l2)) 
        #f
        (fold-right (lambda (val acc) (and acc (set-contains? val s2)))
                    #t
                    lst1))))

(assert-t (set-equal? (list-to-set < (list 1 2 3)) (list-to-set < (list 1 2 3))))
(assert-t (set-equal? (list-to-set < (list 1 2 3)) (list-to-set < (list 3 2 1))))
(assert-f (set-equal? (list-to-set < (list 1 2 3)) (list-to-set < (list 1 2 3 4))))
(assert-f (set-equal? (list-to-set < (list 1 2 3)) (list-to-set < (list 1 2))))
(assert-f (set-equal? (list-to-set < (list )) (list-to-set < (list 1 2 3))))
(assert-t (set-equal? (new-number-set) (new-number-set)))

(define (set-union s1 s2)
  (set-check s1) (set-check s2)
  (fold-right set-add s2 (set-to-list s1)))

(assert-t (set-equal? (set-union (new-number-set) (new-number-set)) (new-number-set)))
(assert-t (set-equal? (set-union (new-number-set) (list-to-set < '(1 1 2))) (list-to-set < '(2 1))))
(assert-t (set-equal? (set-union (list-to-set < '(1 2 3)) (list-to-set < '(1 1 2))) (list-to-set < '(1 2 3))))

(define (set-intersection s1 s2)
  (list-to-set < (filter (lambda (elm) (set-contains? elm s2)) (set-to-list s1))))

(define s1 (new-number-set))
(define s2 (list-to-set < (list 1 1 2)))
(define s3 (list-to-set < (list 1 2 3)))
(define s4 (list-to-set < (list 2 3 4)))
(define s5 (list-to-set < (list 7 8 9)))
(assert-t (set-equal? (set-intersection s1 s1) (new-number-set)))
(assert-t (set-equal? (set-intersection s1 s2) (new-number-set)))
(assert-t (set-equal? (set-intersection s2 s2) s2))
(assert-t (set-equal? (set-intersection s2 s3) (list-to-set < '(1 2))))
(assert-t (set-equal? (set-intersection s3 s4) (list-to-set < '(2 3))))
(assert-t (set-equal? (set-intersection s2 s5) (new-number-set)))

(define (set-diff s1 s2)
  (list-to-set (set-elm-less s1)
   (filter (lambda (elm) (not (set-contains? elm s2)))
	   (set-to-list s1))))
(assert-t (set-equal? (set-diff s1 s1) (new-number-set)))
(assert-t (set-equal? (set-diff s2 s1) s2))
(assert-t (set-equal? (set-diff s3 s2) (list-to-set < '(3))))
(assert-t (set-equal? (set-diff s4 s2) (list-to-set < '(4 3))))
(assert-t (set-equal? (set-diff s5 s2) s5))

